using System;
using System.Collections.Generic;
using System.Text;
using System.Collections;
using gudusoft.gsqlparser;
using gudusoft.gsqlparser.Units;
using System.IO;
using System.Xml.Linq;
using System.Drawing;

namespace ColumnImpact
{
    enum ClauseType
    {
        where, connectby, startwith, orderby, join, undefine, select, groupby
    }

    class TResultEntry
    {
        public Point location;
        public int tokenLen;
        public string targetColumn;
        public TLzTable targetTable;
        public ClauseType clause;
        public TColumn columnObject;

        public TResultEntry(TLzTable table, String viewName, String column, ClauseType clause,
                Point location, int tokenLen)
        {
            this.targetTable = table;
            this.targetColumn = column;
            this.clause = clause;
            this.location = location;
            this.tokenLen = tokenLen;
            columnObject = new TColumn();
            columnObject.columnName = "*";
            columnObject.viewName = viewName;
        }

        public TResultEntry(TLzTable table, TColumn columnObject, String column,
            ClauseType clause, Point location, int tokenLen)
        {
            this.targetTable = table;
            this.targetColumn = column;
            this.clause = clause;
            this.location = location;
            this.tokenLen = tokenLen;
            this.columnObject = columnObject;
        }

    }

    class TAlias
    {
        public string column;
        public string alias;
        public Point location;
        public int tokenLine;
    }

    class TSourceColumn
    {
        public string tableOwner;
        public string tableName;
        public string name;
        public List<ClauseType> clauses = new List<ClauseType>();
        public Hashtable locations = new Hashtable();
        public Hashtable tokenLines = new Hashtable();
    }

    public interface IMetaDatabaseFilter
    {
        bool metaDatabaseTableColumn(string tableOwner, string tableName, string columnName);
    }

    internal class TColumn
    {
        public List<String> tableNames = new List<String>();
        public string columnName;
        public string columnPrex;
        public string orignColumn;
        public Point location;
        public int tokenLine;
        public String viewName;
        public String expression = "";

        public string getFullName(String tableName)
        {
            if (tableName != null)
            {
                return tableName + "." + columnName;
            }
            else
            {
                return columnName;
            }
        }

        public string getOrigName()
        {
            if (columnPrex != null)
            {
                return columnPrex + "." + columnName;
            }
            else
            {
                return columnName;
            }
        }

    }

    class TTable
    {
        public string tableName;
        public string prefixName;
        public string tableAlias;
    }

    internal class Utf8StringWriter : StringWriter
    {
        public Utf8StringWriter(StringBuilder sb) : base(sb) { }

        public override Encoding Encoding { get { return Encoding.UTF8; } }
    }

    class columnsInExpr
    {

        private TLzCustomExpression expr;
        private List<TColumn> columns;
        private int level;
        private columnImpact impact;
        private TCustomSqlStatement stmt;
        private bool collectExpr;

        public columnsInExpr(columnImpact impact, TLzCustomExpression expr, List<TColumn> columns, TCustomSqlStatement stmt, int level,
                bool collectExpr)
        {
            this.stmt = stmt;
            this.impact = impact;
            this.expr = expr;
            this.columns = columns;
            this.level = level;
            this.collectExpr = collectExpr;
        }

        private Boolean searchColumnVisitor(TLz_Node pnode, Boolean pIsLeafNode)
        {
            TLzCustomExpression lcexpr = (TLzCustomExpression)pnode;
            if (lcexpr.oper == TLzOpType.Expr_Attr)
            {
                columns.Add(impact.attrToColumn((TLz_Attr)lcexpr.lexpr, stmt, expr,
                        collectExpr));
            }
            else if (lcexpr.oper == TLzOpType.Expr_FuncCall)
            {
                TLz_FuncCall func = (TLz_FuncCall)lcexpr.lexpr;
                if (func.args != null)
                {
                    for (int k = 0; k < func.args.Count(); k++)
                    {
                        if (func.args[k] is TLzCustomExpression)
                        {
                            TLzCustomExpression expr = (TLzCustomExpression)func.args[k];
                            expr.InOrderTraverse(searchColumnVisitor);
                        }
                    }
                }
                if (func._analyticFunction != null)
                {
                    TLz_List list = func._analyticFunction._partitionExprList;
                    addColumnToList(list, stmt);

                    if (func._analyticFunction._sortclause != null)
                    {
                        list = func._analyticFunction._sortclause.list1;
                        addColumnToList(list, stmt);

                        list = func._analyticFunction._sortclause.list2;
                        addColumnToList(list, stmt);
                    }
                }
                if (func._ndExpr is TLzCustomExpression)
                {
                    ((TLzCustomExpression)func._ndExpr).InOrderTraverse(searchColumnVisitor);
                }
            }
            else if (lcexpr.oper == TLzOpType.Expr_subquery)
            {
                if (lcexpr.lexpr is TSelectSqlStatement)
                {
                    foreach (TLzField field in ((TSelectSqlStatement)lcexpr.lexpr).Fields)
                    {
                        impact.linkFieldToTables(null,
                            field,
                            (TSelectSqlStatement)lcexpr.lexpr,
                            level + 1);
                    }
                }
            }
            else if (lcexpr.oper == TLzOpType.Expr_Case)
            {
                TLzCaseExpression expr = (TLzCaseExpression)lcexpr;
                TLzCustomExpression conditionExpr = expr.ConditionExpr;
                if (conditionExpr != null && !impact.traceView)
                {
                    conditionExpr.InOrderTraverse(searchColumnVisitor);
                }
                TLzCustomExpression defaultExpr = expr.DefaultExpr;
                if (defaultExpr != null)
                {
                    defaultExpr.InOrderTraverse(searchColumnVisitor);
                }
                TLz_List list = expr.WhenThenList;
                addColumnToList(list, stmt);
            }
            return true;
        }

        private void addColumnToList(TLz_List list, TCustomSqlStatement stmt)
        {
            if (list != null)
            {
                for (int i = 0; i < list.Count(); i++)
                {
                    List<TLzCustomExpression> exprList = new List<TLzCustomExpression>();

                    if (list[i] is TLz_SortGroupBy && !impact.traceView)
                    {
                        exprList.Add((TLzCustomExpression)((TLz_SortGroupBy)list[i]).node);
                    }
                    else if (list[i] is TLzCustomExpression)
                    {
                        exprList.Add((TLzCustomExpression)list[i]);
                    }
                    else if (list[i] is TLzCaseWhen)
                    {
                        if (!impact.traceView)
                            exprList.Add(((TLzCaseWhen)list[i]).WhenExpr);
                        exprList.Add(((TLzCaseWhen)list[i]).ThenExpr);
                    }

                    foreach (TLzCustomExpression lcexpr in exprList)
                    {
                        lcexpr.InOrderTraverse(searchColumnVisitor);
                    }
                }
            }
        }

        public void searchColumn()
        {
            this.expr.InOrderTraverse(searchColumnVisitor);
        }
    }


    public class columnImpact
    {
        /*store the dependency relations*/
        Hashtable dependMap = new Hashtable();
        /*store the relations of alias to column*/
        List<TAlias> aliases = new List<TAlias>();
        List<TCustomSqlStatement> stmts = new List<TCustomSqlStatement>();
        String currentSource = null;
        Boolean simply = false;
        Boolean isXML = false;
        Hashtable cteMap = new Hashtable();
        StringBuilder buffer = new StringBuilder();
        Hashtable currentClauseMap = new Hashtable();
        IMetaDatabaseFilter filter;
        int columnNumber = 0;
        TCustomSqlStatement subquery = null;
        string viewName;
        internal bool traceView = false;


        public String ImpactResult
        {
            get { return buffer.ToString(); }
        }


        public columnImpact(String sql, TDbVendor dbVendor, Boolean simply, Boolean isXML)
        {
            this.simply = simply;
            this.isXML = isXML;
            TGSqlParser sqlparser = new TGSqlParser(dbVendor);
            sqlparser.SqlText.Text = sql;
            impactSQL(sqlparser);
        }

        public columnImpact(FileInfo file, TDbVendor dbVendor, Boolean simply, Boolean isXML)
        {
            this.simply = simply;
            this.isXML = isXML;
            TGSqlParser sqlparser = new TGSqlParser(dbVendor);
            sqlparser.Sqlfilename = file.FullName;
            impactSQL(sqlparser);
        }

        public columnImpact(String sql, TDbVendor dbVendor, Boolean simply, Boolean isXML, IMetaDatabaseFilter filter)
        {
            this.simply = simply;
            this.isXML = isXML;
            this.filter = filter;
            TGSqlParser sqlparser = new TGSqlParser(dbVendor);
            sqlparser.SqlText.Text = sql;
            impactSQL(sqlparser);
        }

        public columnImpact(FileInfo file, TDbVendor dbVendor, Boolean simply, Boolean isXML, IMetaDatabaseFilter filter)
        {
            this.simply = simply;
            this.isXML = isXML;
            this.filter = filter;
            TGSqlParser sqlparser = new TGSqlParser(dbVendor);
            sqlparser.Sqlfilename = file.FullName;
            impactSQL(sqlparser);
        }

        public columnImpact(FileInfo file, TDbVendor dbVendor, Boolean simply, Boolean isXML, Boolean traceView, IMetaDatabaseFilter filter)
        {
            if (traceView)
            {
                this.traceView = true;
                this.simply = true;
                this.isXML = isXML;
            }
            else
            {
                this.traceView = false;
                this.simply = simply;
                this.isXML = isXML;
            }
            this.filter = filter;
            TGSqlParser sqlparser = new TGSqlParser(dbVendor);
            sqlparser.Sqlfilename = file.FullName;
            impactSQL(sqlparser);
        }


        private void impactSQL(TGSqlParser sqlparser)
        {
            int ret = sqlparser.Parse();

            if (ret != 0)
            {
                buffer.AppendLine(sqlparser.ErrorMessages);
            }
            else
            {
                XDocument doc = null;
                XElement columnImpactResult = null;
                if (simply && isXML)
                {
                    doc = new XDocument();
                    columnImpactResult = new XElement("columnImpactResult");
                }

                for (int k = 0; k < sqlparser.SqlStatements.Count(); k++)
                {
                    if (sqlparser.SqlStatements[k] is TCustomSqlStatement)
                    {
                        dependMap.Clear();
                        aliases.Clear();
                        currentSource = null;
                        cteMap.Clear();
                        currentClauseMap.Clear();
                        stmts.Clear();
                        TCustomSqlStatement select = (TCustomSqlStatement)sqlparser.SqlStatements[k];
                        if (select.cte != null && select.cte.ctes != null && select.cte.ctes.Count() > 0)
                        {
                            for (int i = 0; i < select.cte.ctes.Count(); i++)
                            {
                                TCommonTableExpression expression = (TCommonTableExpression)select.cte.ctes[i];
                                cteMap.Add(expression.ndTablename.AsText, expression.ndSelect_with_parens);
                            }
                        }

                        columnNumber = 0;
                        impactSqlFromStatement(select);

                        if (traceView && !isXML)
                        {
                            if (select is TCreateViewSqlStatement)
                            {
                                TSelectSqlStatement stmt = ((TCreateViewSqlStatement)select).SubQuery;
                                if (stmt.WhereClause != null)
                                {
                                    buffer.Append("rt=vWhere\tview="
                                            + viewName
                                            + "\twhere=")
                                            .Append(stmt.WhereClause
                                                    .AsText.Replace("\r\n", "\n").Replace("\n", " "))
                                            .Append("\r\n");
                                }
                                if (stmt.Tables != null)
                                {
                                    StringBuilder tableBuffer = new StringBuilder();
                                    List<String> list = new List<String>();
                                    for (int i = 0; i < stmt.Tables.Count(); i++)
                                    {
                                        List<String> tables = new List<String>();
                                        getTableNames(tables,
                                                stmt.Tables[i]);
                                        String[] tableNames = tables.ToArray();
                                        if (tableNames != null)
                                        {
                                            for (int j = 0; j < tableNames.Length; j++)
                                            {
                                                bool exist = false;
                                                for (int z = 0; z < list.Count; z++)
                                                {
                                                    if (list[z].ToUpper()
                                                            .Equals(tableNames[j].ToUpper()))
                                                    {
                                                        exist = true;
                                                        break;
                                                    }
                                                }
                                                if (!exist)
                                                    list.Add(tableNames[j]);
                                            }
                                        }
                                    }

                                    for (int i = 0; i < list.Count; i++)
                                    {
                                        tableBuffer.Append(list[i]);
                                        if (i < list.Count - 1)
                                            tableBuffer.Append(", ");
                                    }
                                    buffer.Append("rt=vTable\tview="
                                            + viewName
                                            + "\ttables=")
                                            .Append(tableBuffer.ToString())
                                            .Append("\r\n");

                                }
                                if (stmt.JoinTables != null)
                                {
                                    for (int i = 0; i < stmt.JoinTables.Count(); i++)
                                    {
                                        if (stmt.JoinTables[i].JoinItems.Count()
                                                 > 0)
                                        {
                                            buffer.Append("rt=vJoin\tview="
                                                    + viewName
                                                    + "\tjoin=")
                                                    .Append(stmt.JoinTables[i].JoinItems.AsText.Replace("\r\n", "\n").Replace("\n", " "))
                                                    .Append("\r\n");
                                        }
                                    }
                                }
                            }

                            Hashtable bufferMap = new Hashtable();
                            Hashtable exprMap = new Hashtable();

                            foreach (TAlias alias in aliases)
                            {
                                if (dependMap.ContainsKey(alias.alias))
                                {
                                    List<TResultEntry> results = (List<TResultEntry>)dependMap[alias.alias];
                                    List<String> nullRealColumns = new List<String>();
                                    foreach (TResultEntry result in results)
                                    {
                                        TColumn columnObject = result.columnObject;
                                        if (columnObject == null
                                                || columnObject.viewName == null)
                                            continue;

                                        if (result.clause != ClauseType.select)
                                            continue;
                                        String column = null;
                                        if (result.columnObject.columnName != null)
                                        {
                                            if ("*".Equals(result.targetColumn))
                                            {
                                                column = result.targetTable.TableFullname.ToLower();
                                            }
                                            else
                                            {
                                                column = (result.targetTable.TableFullname + "." + result.targetColumn).ToLower();
                                            }
                                        }
                                        else
                                        {
                                            if (nullRealColumns.Contains(result.columnObject.expression))
                                                continue;
                                            else
                                                nullRealColumns.Add(result.columnObject.expression);

                                        }


                                        String temp = ("rt=col\tview="
                                            + columnObject.viewName
                                            + "\t"
                                            + "column="
                                            + (alias.alias.IndexOf(".") != -1 ? alias.alias.Substring(alias.alias.LastIndexOf(".") + 1)
                                                    : alias.alias)
                                            + "\t"
                                            + (column != null ? ("source="
                                                    + column + "\t") : "") + "expression=");

                                        if (!bufferMap.ContainsKey(temp.ToUpper()))
                                        {
                                            bufferMap.Add(temp.ToUpper(),
                                                    temp);
                                        }
                                        if (columnObject.expression != null
                                                && columnObject.expression.Trim().Length
                                                         > 0)
                                        {
                                            if (!exprMap.ContainsKey(temp.ToUpper()))
                                            {
                                                List<String> exprList = new List<String>();
                                                exprList.Add(columnObject.expression);
                                                exprMap.Add(temp.ToUpper(),
                                                        exprList);
                                            }
                                            else
                                            {
                                                List<String> exprList = (List<String>)exprMap[temp.ToUpper()];
                                                bool exist = false;
                                                for (int i = 0; i < exprList.Count; i++)
                                                {
                                                    if (exprList[i].ToUpper().Equals(columnObject.expression.ToUpper()))
                                                    {
                                                        exist = true;
                                                        break;
                                                    }
                                                }
                                                if (!exist)
                                                    exprList.Add(columnObject.expression);
                                            }
                                        }
                                    }
                                }
                            }

                            IEnumerator iter = bufferMap.Values.GetEnumerator();
                            while (iter.MoveNext())
                            {
                                String temp = (String)iter.Current;
                                buffer.Append(temp);
                                List<String> exprList = (List<String>)exprMap[temp.ToUpper()];
                                String expr = "";
                                if (exprList != null)
                                {
                                    for (int i = 0; i < exprList.Count; i++)
                                    {

                                        expr += exprList[i];
                                        if (i < exprList.Count - 1)
                                            expr += "; ";
                                    }
                                }
                                buffer.Append(expr + "\r\n");
                            }
                        }
                        else if (simply)
                        {
                            if (!isXML)
                            {
                                foreach (TAlias alias in aliases)
                                {

                                    buffer.Append(alias.alias + " depends on: ");

                                    List<String> collections = new List<String>();

                                    if (dependMap.ContainsKey(alias.alias))
                                    {
                                        List<TResultEntry> results = (List<TResultEntry>)dependMap[alias.alias];
                                        foreach (TResultEntry result in results)
                                        {
                                            string column = null;
                                            if (traceView
                                                    && result.clause != ClauseType.select)
                                                continue;
                                            if ("*".Equals(result.targetColumn))
                                            {
                                                column = result.targetTable.TableFullname.ToLower();
                                            }
                                            else
                                            {
                                                column = (result.targetTable.TableFullname + "." + result.targetColumn).ToLower();
                                            }
                                            if (!collections.Contains(column)) collections.Add(column);
                                        }
                                    }

                                    List<String> list = new List<String>(collections);
                                    for (int i = 0; i < list.Count; i++)
                                    {
                                        if (i < collections.Count - 1)
                                            buffer.Append(list[i] + ", ");
                                        else
                                            buffer.Append(list[i]);
                                    }

                                    buffer.AppendLine();

                                }
                            }
                            else
                            {

                                foreach (TAlias alias in aliases)
                                {
                                    XElement targetColumn = new XElement("targetColumn");
                                    if (!alias.alias.Equals(alias.column))
                                        targetColumn.Add(new XAttribute("alias", alias.alias));
                                    targetColumn.Add(new XAttribute("coordinate", alias.location.X + "," + alias.location.Y));
                                    targetColumn.Add(new XAttribute("tokenLine", alias.tokenLine));
                                    targetColumn.Add(new XAttribute("name", alias.column));

                                    columnImpactResult.Add(targetColumn);

                                    Hashtable collections = new Hashtable();
                                    List<string> keyList = new List<string>();

                                    if (dependMap.ContainsKey(alias.alias))
                                    {
                                        List<TResultEntry> results = (List<TResultEntry>)dependMap[alias.alias];
                                        foreach (TResultEntry result in results)
                                        {
                                            String key = null;
                                            if (traceView
                                                   && result.clause != ClauseType.select)
                                                continue;
                                            if ("*".Equals(result.targetColumn))
                                            {
                                                key = result.targetTable.TableFullname.ToLower();
                                            }
                                            else
                                            {
                                                key = (result.targetTable.TableFullname + "." + result.targetColumn).ToLower();
                                            }

                                            TSourceColumn sourceColumn = null;
                                            if (collections.ContainsKey(key))
                                            {
                                                sourceColumn = (TSourceColumn)collections[key];
                                                if (!sourceColumn.clauses.Contains(result.clause))
                                                {
                                                    sourceColumn.clauses.Add(result.clause);
                                                }
                                                if (result.location != null)
                                                {
                                                    if (!sourceColumn.locations.ContainsKey(result.clause))
                                                        sourceColumn.locations.Add(result.clause,
                                                                new List<Point>());
                                                    List<Point> ys = (List<Point>)sourceColumn.locations[result.clause];
                                                    if (!ys.Contains(result.location))
                                                        ys.Add(result.location);
                                                }
                                                if (result.tokenLen != 0)
                                                {
                                                    if (!sourceColumn.tokenLines.ContainsKey(result.clause))
                                                        sourceColumn.tokenLines.Add(result.clause,
                                                                new List<int>());
                                                    List<int> ys = (List<int>)sourceColumn.tokenLines[result.clause];
                                                    if (!ys.Contains(result.tokenLen))
                                                        ys.Add(result.tokenLen);
                                                }

                                            }
                                            else
                                            {
                                                sourceColumn = new TSourceColumn();
                                                collections.Add(key, sourceColumn);
                                                keyList.Add(key);
                                                sourceColumn.tableOwner = result.targetTable.TableOwner;
                                                sourceColumn.tableName = result.targetTable.TableName;
                                                if (!"*".Equals(result.targetColumn))
                                                {
                                                    sourceColumn.name = result.targetColumn;
                                                }
                                                if (!sourceColumn.clauses.Contains(result.clause))
                                                {
                                                    sourceColumn.clauses.Add(result.clause);
                                                }
                                                if (result.location != null)
                                                {
                                                    if (!sourceColumn.locations.ContainsKey(result.clause))
                                                        sourceColumn.locations.Add(result.clause,
                                                                new List<Point>());
                                                    List<Point> ys = (List<Point>)sourceColumn.locations[result.clause];
                                                    if (!ys.Contains(result.location))
                                                        ys.Add(result.location);
                                                }
                                                if (result.tokenLen != 0)
                                                {
                                                    if (!sourceColumn.tokenLines.ContainsKey(result.tokenLen))
                                                        sourceColumn.tokenLines.Add(result.clause,
                                                                new List<int>());
                                                    List<int> ys = (List<int>)sourceColumn.tokenLines[result.clause];
                                                    if (!ys.Contains(result.tokenLen))
                                                        ys.Add(result.tokenLen);
                                                }
                                            }
                                        }

                                        foreach (string key in keyList)
                                        {
                                            TSourceColumn sourceColumn = (TSourceColumn)collections[key];
                                            if (sourceColumn.clauses.Count > 0)
                                            {
                                                for (int j = 0; j < sourceColumn.clauses.Count; j++)
                                                {
                                                    ClauseType clause = sourceColumn.clauses[j];
                                                    XElement element = new XElement("sourceColumn");
                                                    {
                                                        StringBuilder buffer = new StringBuilder();
                                                        switch (clause)
                                                        {
                                                            case ClauseType.where:
                                                                buffer.Append("where");
                                                                break;
                                                            case ClauseType.connectby:
                                                                buffer.Append("connect by");
                                                                break;
                                                            case ClauseType.startwith:
                                                                buffer.Append("start with");
                                                                break;
                                                            case ClauseType.orderby:
                                                                buffer.Append("order by");
                                                                break;
                                                            case ClauseType.join:
                                                                buffer.Append("join");
                                                                break;
                                                            case ClauseType.select:
                                                                buffer.Append("select");
                                                                break;
                                                            case ClauseType.groupby:
                                                                buffer.Append("group by");
                                                                break;
                                                        }
                                                        if (buffer.ToString().Trim().Length > 0)
                                                        {
                                                            XAttribute attribute = new XAttribute("clause", buffer.ToString());
                                                            element.Add(attribute);
                                                        }
                                                    }
                                                    {
                                                        StringBuilder buffer = new StringBuilder();
                                                        buildLocationString(sourceColumn,
                                                                clause,
                                                                buffer);
                                                        if (buffer.ToString().Trim().Length > 0)
                                                        {
                                                            XAttribute attribute = new XAttribute("coordinate", buffer.ToString());
                                                            element.Add(attribute);
                                                        }

                                                    }
                                                    {
                                                        StringBuilder buffer = new StringBuilder();
                                                        buildTokenLineString(sourceColumn,
                                                                clause,
                                                                buffer);
                                                        if (buffer.ToString().Trim().Length > 0)
                                                        {
                                                            XAttribute attribute = new XAttribute("tokenLine", buffer.ToString());
                                                            element.Add(attribute);
                                                        }

                                                    }
                                                    if (sourceColumn.name != null && sourceColumn.name.Trim().Length > 0)
                                                    {
                                                        XAttribute attribute = new XAttribute("name", sourceColumn.name);
                                                        element.Add(attribute);
                                                    }
                                                    if (sourceColumn.tableName != null && sourceColumn.tableName.Trim().Length > 0)
                                                    {
                                                        XAttribute attribute = new XAttribute("tableName", sourceColumn.tableName);
                                                        element.Add(attribute);
                                                    }
                                                    if (sourceColumn.tableOwner != null && sourceColumn.tableOwner.Trim().Length > 0)
                                                    {
                                                        XAttribute attribute = new XAttribute("tableOwner", sourceColumn.tableOwner);
                                                        element.Add(attribute);
                                                    }
                                                    targetColumn.Add(element);
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }

                if (doc != null)
                {
                    doc.Add(columnImpactResult);
                    using (StringWriter writer = new Utf8StringWriter(buffer))
                    {
                        doc.Save(writer, SaveOptions.None);
                    }
                }
            }
        }

        private void getTableNames(List<String> tableNames, TLzTable table)
        {
            if (table.SubQuery != null)
            {
                for (int i = 0; i < table.SubQuery.Tables.Count(); i++)
                {
                    getTableNames(tableNames,
                            table.SubQuery.Tables[i]);
                }
            }
            else
            {
                tableNames.Add(table.TableFullname);
            }
        }

        private void buildLocationString(TSourceColumn sourceColumn,
        ClauseType clauseType, StringBuilder locationBuffer)
        {
            List<Point> ys = (List<Point>)sourceColumn.locations[clauseType];
            if (ys != null)
            {
                for (int z = 0; z < ys.Count; z++)
                {
                    locationBuffer.Append(ys[z].X + "," + ys[z].Y);
                    if (z < ys.Count - 1)
                        locationBuffer.Append(";");
                }
            }
        }

        private void buildTokenLineString(TSourceColumn sourceColumn,
        ClauseType clauseType, StringBuilder tokenLineBuffer)
        {
            List<int> ys = (List<int>)sourceColumn.tokenLines[clauseType];
            if (ys != null)
            {
                for (int z = 0; z < ys.Count; z++)
                {
                    tokenLineBuffer.Append(ys[z]);
                    if (z < ys.Count - 1)
                        tokenLineBuffer.Append(";");
                }
            }
        }

        private void impactSqlFromStatement(TCustomSqlStatement select)
        {
            
            
            if (select is TSelectSqlStatement)
            {
                TSelectSqlStatement stmt = (TSelectSqlStatement)select;
                if (stmt.SelectSetType != TSelectSetType.sltNone)
                {
                    impactSqlFromStatement(stmt.LeftStmt);
                    impactSqlFromStatement(stmt.RightStmt);
                }
                else
                {
                    if ( !stmts.Contains(select) && isTopLevel(select) )
                        stmts.Add(select);
                    else return;
                    
                    foreach (TLzField field in select.Fields)
                    {
                        linkFieldToTables(null, field, select, 0);
                    }
                }
            }
            else if (select is TInsertSqlStatement && ((TInsertSqlStatement)select).subquery != null)
            {
                impactSqlFromStatement(((TInsertSqlStatement)select).subquery);
            }
            else if (select is TCreateViewSqlStatement && ((TCreateViewSqlStatement)select).SubQuery != null)
            {
                viewName = ((TCreateViewSqlStatement)select).ViewName;
                impactSqlFromStatement(((TCreateViewSqlStatement)select).SubQuery);
            }
            else if (select.Fields != null && select.Fields.Count()>0)
            {
                if (select is TInsertSqlStatement)
                    return;

                if (!stmts.Contains(select) && isTopLevel(select))
                    stmts.Add(select);
                else return;

                foreach (TLzField field in select.Fields)
                {
                    linkFieldToTables(null, field, select, 0);
                }
            }
            else if (select.ChildNodes != null && select.ChildNodes.Count() > 0)
            {
                for (int i = 0; i < select.ChildNodes.Count(); i++)
                {
                    Object node = select.ChildNodes[i];
                    if (node is TCustomSqlStatement)
                    {
                        impactSqlFromStatement(node as TCustomSqlStatement);
                    }
                }
            }
        }

        class MetaDatabaseFilter : IMetaDatabaseFilter
        {
            public bool metaDatabaseTableColumn(string tableOwner, string tableName, string columnName)
            {
                if ("other_table".Equals(tableName))
                {
                    if ("c1".Equals(columnName))
                        return true;
                    else return false;
                }
                if ("some_table".Equals(tableName))
                {
                    if ("c1".Equals(columnName))
                        return false;
                    else return true;
                }
                return true;
            }
        }

        static void Main(string[] args)
        {
            if (args.Length == 0)
            {
                Console.WriteLine("Usage: {0} scriptfile [/d]/[/s [/xml] [/c]]/[/v] [/o <output file path>] [/t <database type>]", Environment.GetCommandLineArgs()[0]);
                Console.WriteLine("/s: Option, display the analysis result simply.");
                Console.WriteLine("/c: Option, display the analysis result simply in column level.");
                Console.WriteLine("/d: Option, display the analysis result in detail.");
                Console.WriteLine("/xml: Option, export the analysis results to XML format, it's valid only if /s is specified");
                Console.WriteLine("/v: Option, trace data lineage in views.");
                Console.WriteLine("/o: Option, write the output stream to the specified file.");
                Console.WriteLine("/t: Option, set the database type. Support oracle,mssql,mysql and db2, the default type is oracle");
                //Console.Read();
                return;
            }


            Array array = Array.CreateInstance(typeof(string), args.Length);

            for (int i = 0; i < array.Length; i++)
            {
                array.SetValue(args[i], i);
            }

            bool traceView = Array.IndexOf(array, "/v") != -1;

            bool simply = traceView || Array.IndexOf(array, "/s") != -1;

            bool isXML = simply && Array.IndexOf(array, "/xml") != -1;

            string outputFile = null;

            int index = Array.IndexOf(array, "/o");

            if (index != -1 && args.Length > index + 1)
            {
                outputFile = args[index + 1];
            }

            StreamWriter writer = null;
            if (outputFile != null)
            {
                try
                {
                    writer = new StreamWriter(outputFile);
                    Console.SetOut(writer);
                }
                finally { }
            }

            TDbVendor vendor = TDbVendor.DbVOracle;

            index = Array.IndexOf(array, "/t");

            if (index != -1 && args.Length > index + 1)
            {
                if (args[index + 1].Equals("mssql"))
                {
                    vendor = TDbVendor.DbVMssql;
                }
                else if (args[index + 1].Equals("db2"))
                {
                    vendor = TDbVendor.DbVDB2;
                }
                else if (args[index + 1].Equals("mysql"))
                {
                    vendor = TDbVendor.DbVMysql;
                }
            }


            columnImpact impact = new columnImpact(new FileInfo(args[0]), vendor, simply, isXML, traceView, new MetaDatabaseFilter());
            Console.Write(impact.ImpactResult);

            if (!simply)
            {
                Console.WriteLine(System.Environment.NewLine + "You can add /s directive to display the analysis result in a simple format.");
            }

            //if (outputFile == null)
            //{
            //    Console.Read();
            //}
            //else
            {
                if (writer != null)
                {
                    writer.Close();
                }
            }

        } //main

        bool findColumnInTables(TColumn column, TCustomSqlStatement select, int level, string columnName,
            Object originLocation)
        {
            bool ret = false;
            foreach (String tableName in column.tableNames)
            {
                if (columnName != null && filter != null)
                {
                    int dotIndex = tableName.LastIndexOf(".");
                    string tableOwner = null;
                    string tableRealName = null;
                    if (dotIndex >= 0)
                    {
                        tableOwner = tableName.Substring(0, dotIndex);
                        tableRealName = tableName.Replace(tableOwner + ".", "");
                    }
                    else
                    {
                        tableRealName = tableName;
                    }
                    if (filter.metaDatabaseTableColumn(tableOwner, tableRealName, columnName))
                    {
                        column.columnName = columnName;
                        if (originLocation != null)
                            column.location = (Point)originLocation;
                        //column.orignColumn = "*";
                        ret |= findColumnInTables(column, tableName, select, level);
                    }
                    else ret |= false;
                }
                else ret |= findColumnInTables(column, tableName, select, level);
            }
            return ret;
        }

        bool findColumnInTables(TColumn column, String tableName, TCustomSqlStatement select, int level)
        {
            return findColumnInTables(column, tableName, select, level, ClauseType.undefine);
        }

        bool findColumnInTables(TColumn column, String tableName, TCustomSqlStatement select, int level, ClauseType clause)
        {
            bool ret = false;
            TLzTableList tables = select.Tables;

            if (tables.Count() == 1)
            {
                TLzTable lzTable = tables[0];
                //buffer.AppendLine(lzTable.AsText);
                if ((lzTable.TableType == TLzTableType.lttAttr)
                    && (tableName == null
                            || (tableName != null
                                    && lzTable.AliasClause == null && lzTable.TableName
                                    .Equals(tableName, StringComparison.OrdinalIgnoreCase)) || (tableName != null
                            && lzTable.AliasClause != null && lzTable.AliasClause.aliastext.Equals(tableName, StringComparison.OrdinalIgnoreCase))))
                {
                    ret = true;

                    if (!simply)
                    {
                        buffer.AppendLine(new string(' ', level) + "--> " + lzTable.AsText + "." + column.columnName);
                    }
                    if (cteMap.ContainsKey(lzTable.Name))
                    {
                        if (!simply)
                        {
                            buffer.AppendLine(new string(' ', level) + "--> WITH CTE");
                        }
                        ret = findColumnInSubQuery((TSelectSqlStatement)cteMap[lzTable.Name], column.columnName, level,
                            column.location);
                    }
                    else
                    {
                        if (currentSource != null && dependMap.ContainsKey(currentSource) && dependMap[currentSource] is List<TResultEntry>)
                        {
                            TCustomSqlStatement stmt = containClasuse(currentClauseMap, select);
                            if (stmt != null)
                            {
                                ((List<TResultEntry>)dependMap[currentSource]).Add(new TResultEntry(lzTable, column, column.columnName, (ClauseType)currentClauseMap[stmt], column.location, column.tokenLine));
                            }
                            else if (select is TSelectSqlStatement)
                            {
                                if (ClauseType.undefine.Equals(clause))
                                    ((List<TResultEntry>)dependMap[currentSource]).Add(new TResultEntry(lzTable, column, column.columnName, ClauseType.select, column.location, column.tokenLine));
                                else
                                    ((List<TResultEntry>)dependMap[currentSource]).Add(new TResultEntry(lzTable, column, column.columnName, clause, column.location, column.tokenLine));
                            }
                            else
                            {
                                ((List<TResultEntry>)dependMap[currentSource]).Add(new TResultEntry(lzTable, column, column.columnName, ClauseType.undefine, column.location, column.tokenLine));
                            }
                        }
                    }
                }
                else if (lzTable.TableType == TLzTableType.lttSubquery)
                {

                    ret = findColumnInSubQuery((TSelectSqlStatement)lzTable.SubQuery, column.columnName, level,
                            column.location);
                }
                else if (select.OwnerStmt is TSelectSqlStatement)
                {
                    subquery = select;
                    ret = findColumnInTables(column,
                            tableName,
                            (TSelectSqlStatement)select.OwnerStmt,
                            level,
                            clause);
                    subquery = null;
                }
            }

            if (ret) return ret;

            foreach (TLzTable lztable in tables)
            {
                switch (lztable.TableType)
                {
                    case TLzTableType.lttAttr:
                        TTable table = TLzTaleToTable(lztable);
                        String alias = table.tableAlias;
                        if (alias != null) alias = alias.Trim();
                        if ((tableName != null) &&
                            ((tableName.Equals(alias, StringComparison.OrdinalIgnoreCase)) || (tableName.Equals(table.tableName, StringComparison.OrdinalIgnoreCase)))
                           )
                        {
                            if (!simply)
                            {
                                buffer.AppendLine(new string(' ', level) + "--> " + table.tableName + "." + column.columnName);
                            }
                            if (cteMap.ContainsKey(lztable.Name))
                            {
                                if (!simply)
                                {
                                    buffer.AppendLine(new string(' ', level) + "--> WITH CTE");
                                }
                                ret = findColumnInSubQuery((TSelectSqlStatement)cteMap[lztable.Name], column.columnName, level,
                                    column.location);
                            }
                            else
                            {
                                if (dependMap.ContainsKey(currentSource) && dependMap[currentSource] is List<TResultEntry>)
                                {
                                    String columnName = column.orignColumn;
                                    if ("*".Equals(columnName))
                                        columnName = column.columnName;
                                    if (currentClauseMap.ContainsKey(select))
                                    {
                                        ((List<TResultEntry>)dependMap[currentSource]).Add(new TResultEntry(lztable, column, columnName, (ClauseType)currentClauseMap[select], column.location, column.tokenLine));
                                    }
                                    else if (select is TSelectSqlStatement)
                                    {
                                        ((List<TResultEntry>)dependMap[currentSource]).Add(new TResultEntry(lztable, column, columnName, ClauseType.select, column.location, column.tokenLine));
                                    }
                                    else
                                    {
                                        ((List<TResultEntry>)dependMap[currentSource]).Add(new TResultEntry(lztable, column, columnName, ClauseType.undefine, column.location, column.tokenLine));
                                    }
                                }
                                ret = true;
                            }
                        }
                        break;
                    case TLzTableType.lttSubquery:
                        for (int i = 0; i < column.tableNames.Count; i++)
                        {
                            String name = column.tableNames[i];
                            TSelectSqlStatement selectStat = (TSelectSqlStatement)lztable.SubQuery;

                            if (selectStat == subquery)
                                continue;

                            if (name == null && column.columnName != null)
                            {
                                ret = findColumnInSubQuery(selectStat, column.columnName, level,
                                    column.location);
                                break;
                            }

                            if (lztable.AliasClause != null && lztable.AliasClause.AsText.Trim().Equals(name, StringComparison.OrdinalIgnoreCase))
                            {
                                ret = findColumnInSubQuery(selectStat, column.columnName, level,
                                    column.location);
                                break;
                            }

                            bool flag = false;
                            for (int j = 0; j < selectStat.Tables.Count(); j++)
                            {
                                if (selectStat.Tables[j].AliasClause != null)
                                {
                                    if (selectStat.Tables[j].TableAlias.Equals(name, StringComparison.OrdinalIgnoreCase))
                                    {
                                        ret = findColumnInSubQuery(selectStat, column.columnName, level,
                                    column.location);
                                        flag = true;
                                        break;
                                    }
                                }
                                else
                                {
                                    if (selectStat.Tables[j].TableName.Equals(name, StringComparison.OrdinalIgnoreCase))
                                    {
                                        ret = findColumnInSubQuery(selectStat, column.columnName, level,
                                    column.location);
                                        flag = true;
                                        break;
                                    }
                                }
                            }
                            if (flag) break;
                        }
                        break;
                    default:
                        break;
                }
                if (ret) break;
            }

            if (!ret && select.OwnerStmt is TSelectSqlStatement)
            {
                subquery = select;
                ret = findColumnInTables(column,
                        tableName,
                        (TSelectSqlStatement)select.OwnerStmt,
                        level,
                        clause);
                subquery = null;
            }

            return ret;
        }

        private TCustomSqlStatement containClasuse(Hashtable currentClauseMap, TCustomSqlStatement select)
        {
            if (currentClauseMap.Contains(select))
                return select;
            else if (select.OwnerStmt is TCustomSqlStatement)
            {
                return containClasuse(currentClauseMap, (TCustomSqlStatement)select.OwnerStmt);
            }
            else return null;
        }

        internal bool findColumnInSubQuery(TSelectSqlStatement select, string columnName, int level, Point originLocation)
        {
            bool ret = false;
            if (select.SelectSetType != TSelectSetType.sltNone)
            {
                bool left = findColumnInSubQuery(select.LeftStmt, columnName, level, originLocation);
                bool right = findColumnInSubQuery(select.RightStmt, columnName, level, originLocation);
                ret = left && right;
            }
            else
            {
                // check colum name in select list of subquery
                TLzField columnField = null;
                if (!"*".Equals(columnName))
                {
                    foreach (TLzField field in select.Fields)
                    {
                        if (field.FieldAlias.Length > 0)
                        {
                            if (field.FieldAlias.Equals(columnName, StringComparison.OrdinalIgnoreCase))
                            {
                                columnField = field;
                                break;
                            }
                        }
                        else
                        {
                            if (field.FieldType == TLzFieldType.lftAttr)
                            {
                                TColumn column = attrToColumn(field.FieldAttr, select);
                                if (columnName != null && columnName.Equals(column.columnName, StringComparison.OrdinalIgnoreCase))
                                {
                                    columnField = field;
                                    break;
                                }
                            }
                        }
                    }
                }
                foreach (TLzField field in select.Fields)
                {
                    if (columnField != null && !field.Equals(columnField))
                    {
                        continue;
                    }
                    if (field.FieldAlias.Length > 0)
                    {
                        ret = "*".Equals(columnName) || field.FieldAlias.Equals(columnName, StringComparison.OrdinalIgnoreCase);
                        if (ret)
                        {
                            // let's check where this column come from?
                            if (!simply)
                            {
                                buffer.AppendLine(new string(' ', level) + "--> " + field.FieldAlias + "(alias)");
                            }
                            linkFieldToTables(null, field, select, level);
                        }
                    }
                    else
                    {
                        if (field.FieldType == TLzFieldType.lftAttr)
                        {
                            TColumn column = attrToColumn(field.FieldAttr, select);
                            ret = "*".Equals(columnName) || columnName.Equals(column.columnName, StringComparison.OrdinalIgnoreCase);
                            if (ret || "*".Equals(column.columnName))
                            {
                                findColumnInTables(column, select, level, ret == false ? columnName : null,
                                    originLocation);
                                findColumnsFromClauses(select, level + 1);
                            }
                        }
                    }

                    if (ret && !"*".Equals(columnName)) break;
                }
            }
            return ret;
        } // findColumnInSubQuery

        internal bool linkFieldToTables(TAlias parentAlias, TLzField field, TCustomSqlStatement select, int level)
        {
            bool ret = false;
            //TLzTableList tables = select.Tables;
            // all items in select list was represented by a TLzField Objects
            switch (field.FieldType)
            {
                case TLzFieldType.lftAttr:
                    TColumn column = attrToColumn(field.FieldAttr, select);
                    Boolean isPseudocolumn = select.DBVendor == TDbVendor.DbVOracle
                        && this.isPseudocolumn(column.columnName);
                    if (level == 0 || parentAlias != null)
                    {
                        TAlias alias = null;
                        if (parentAlias != null)
                        {
                            alias = parentAlias;
                        }
                        else
                        {
                            alias = new TAlias();
                            alias.column = field.AsText;
                            alias.alias = field.AsText;
                            alias.location = new Point(field.StartToken.XPosition, field.StartToken.YPosition);
                            alias.tokenLine = field.StartToken.TokenLen;
                            if (field.aliasclause != null)
                            {
                                alias.alias = field.FieldAlias;
                                alias.column = field.FieldFullname;
                                alias.location = new Point(field.aliasclause.AliasToken.XPosition, field.aliasclause.AliasToken.YPosition);
                                alias.tokenLine = field.aliasclause.AliasToken.TokenLen;
                            }
                            aliases.Add(alias);
                        }
                        currentSource = alias.alias;
                        if (!dependMap.ContainsKey(currentSource))
                            dependMap.Add(currentSource, new List<TResultEntry>());

                        if (!simply && parentAlias == null)
                        {
                            if (!alias.alias.Equals(column.getOrigName()))
                            {
                                buffer.AppendLine(System.Environment.NewLine + "Search " + alias.alias + (level == 0 ? (" <<column_"
                                            + (++columnNumber) + ">>") : ""));
                                buffer.AppendLine("--> " + column.getOrigName() + (!isPseudocolumn
                                            && column.tableNames.Count > 1 ? (" <<GUESS>>")
                                            : ""));
                            }
                            else
                            {
                                buffer.AppendLine(System.Environment.NewLine + "Search " + column.getOrigName() + (level == 0 ? (" <<column_"
                                            + (++columnNumber) + ">>") : "") + (!isPseudocolumn
                                            && column.tableNames.Count > 1 ? (" <<GUESS>>")
                                            : ""));
                                level -= 1;
                            }
                        }

                    }
                    if (isPseudocolumn)
                    {
                        break;
                    }
                    ret = findColumnInTables(column, select, level + 1, null, null);
                    findColumnsFromClauses(select, level + 2);
                    break;
                case TLzFieldType.lftExpression:
                    if (level == 0)
                    {
                        TAlias alias = new TAlias();

                        if (select is TUpdateSqlStatement)
                        {
                            alias.column = field.FieldExpr.lexpr.AsText;
                            alias.alias = alias.column;
                            TLz_Node expr = field.FieldExpr.lexpr;
                            alias.location = new Point(expr.StartToken.XPosition, expr.StartToken.YPosition);
                            alias.tokenLine = expr.StartToken.TokenLen;
                        }
                        else
                        {
                            alias.column = field.AsText;
                            alias.alias = alias.column;
                            alias.location = new Point(field.StartToken.XPosition, field.StartToken.YPosition);
                            alias.tokenLine = field.StartToken.TokenLen;
                        }
                        if (field.aliasclause != null)
                        {
                            alias.alias = field.FieldAlias;
                            alias.column = field.FieldFullname;
                            alias.location = new Point(field.aliasclause.AliasToken.XPosition, field.aliasclause.AliasToken.YPosition);
                            alias.tokenLine = field.aliasclause.AliasToken.TokenLen;
                        }
                        aliases.Add(alias);
                        if (!simply)
                        {
                            buffer.AppendLine(System.Environment.NewLine + "Search " + alias.alias + (level == 0 ? (" <<column_"
                                            + (++columnNumber) + ">>") : ""));
                        }

                        currentSource = alias.alias;
                        if (!dependMap.ContainsKey(currentSource))
                            dependMap.Add(currentSource, new List<TResultEntry>());
                    }

                    List<TColumn> columns = exprToColumn(field.FieldExpr, select, level, true);

                    if (columns.Count == 0)
                    {
                        TColumn nullColumn = new TColumn();
                        nullColumn.expression = field.FieldExpr.AsText;
                        nullColumn.viewName = this.viewName;
                        TLzTableList tables = select.Tables;
                        for (int i = 0; i < tables.Count(); i++)
                        {
                            TLzTable lztable = tables[i];
                            TTable table = TLzTaleToTable(lztable);
                            if (!nullColumn.tableNames.Contains(table.tableName))
                            {
                                nullColumn.tableNames.Add(table.tableName);
                            }
                        }
                        columns.Add(nullColumn);
                    }

                    if (select is TUpdateSqlStatement)
                    {
                        while (columns.Count > 1)
                        {
                            columns.RemoveAt(columns.Count - 1);
                        }
                    }
                    if (!simply)
                    {
                        foreach (TColumn column1 in columns)
                        {
                            if (level == 0)
                            {
                                buffer.AppendLine(new string(' ', level) + "--> " + column1.getOrigName());
                            }
                        }
                    }

                    foreach (TColumn column1 in columns)
                    {
                        if (level == 0)
                        {
                            if (!simply)
                            {
                                buffer.AppendLine(System.Environment.NewLine + "Search " + column1.getOrigName());
                            }
                        }
                        if (!(select is TUpdateSqlStatement) && column1.columnName != null)
                        {
                            findColumnInTables(column1, select, level + 1,
                                null,
                                null);
                        }
                        findColumnsFromClauses(select, level + 2);
                    }


                    if (field.FieldExpr.oper == TLzOpType.Expr_FuncCall)
                    {
                        TLz_FuncCall func = (TLz_FuncCall)field.FieldExpr.lexpr;
                        // buffer.AppendLine("function name {0}", func.funcname.AsText);
                        if ((func.funcname.AsText.Equals("count", StringComparison.OrdinalIgnoreCase))
                            || (func.funcname.AsText.Equals("sum", StringComparison.OrdinalIgnoreCase))
                            || (func.funcname.AsText.Equals("row_number", StringComparison.OrdinalIgnoreCase))
                            )
                        {
                            if (!simply)
                            {
                                buffer.AppendLine(new string(' ', level + 1) + "--> aggregate function " + func.AsText);
                                for (int i = 0; i < select.Tables.Count(); i++)
                                {
                                    buffer.AppendLine(new string(' ', level + 1) + "--> table " + select.Tables[i].AsText);
                                }
                            }
                            // check column in function arguments
                            if (func.args != null)
                            {
                                for (int k = 0; k < func.args.Count(); k++)
                                {
                                    if (func.args[k] is TLzCustomExpression)
                                    {
                                        TLzCustomExpression expr = (TLzCustomExpression)func.args[k];
                                        List<TColumn> columns1 = exprToColumn(expr, select, level + 1);
                                        foreach (TColumn column1 in columns1)
                                        {
                                            findColumnInTables(column1, select, level + 1,
                                            null,
                                            null);
                                            findColumnsFromClauses(select, level + 2);
                                        }
                                    }
                                }
                            }
                            else
                            {
                                if (dependMap.ContainsKey(currentSource) && dependMap[currentSource] is List<TResultEntry>)
                                {
                                    Point point = new Point((int)func.EndToken.XPosition, (int)func.EndToken.YPosition);
                                    int tokenLine = func.EndToken.TokenLen;
                                    if (func._stArgPrefix != null)
                                    {
                                        point = new Point((int)func._stArgPrefix.XPosition, (int)func._stArgPrefix.YPosition);
                                        tokenLine = func._stArgPrefix.TokenLen;
                                    }
                                    if (currentClauseMap.ContainsKey(select))
                                    {
                                        ((List<TResultEntry>)dependMap[currentSource]).Add(new TResultEntry(select.Tables[0], viewName, "*", (ClauseType)currentClauseMap[select], point, tokenLine));
                                    }
                                    else if (select is TSelectSqlStatement)
                                    {
                                        ((List<TResultEntry>)dependMap[currentSource]).Add(new TResultEntry(select.Tables[0], viewName, "*", ClauseType.select, point, tokenLine));
                                    }
                                    else
                                    {
                                        ((List<TResultEntry>)dependMap[currentSource]).Add(new TResultEntry(select.Tables[0], viewName, "*", ClauseType.undefine, point, tokenLine));
                                    }
                                }
                            }

                            if (func._analyticFunction != null)
                            {
                                TLz_List list = func._analyticFunction._partitionExprList;

                                findColumnsFromList(select, level, list);


                                if (func._analyticFunction._sortclause != null)
                                {
                                    list = func._analyticFunction._sortclause.list1;

                                    findColumnsFromList(select, level + 1, list);


                                    list = func._analyticFunction._sortclause.list2;

                                    findColumnsFromList(select, level + 1, list);

                                }
                            }


                            findColumnsFromClauses(select, level + 2);


                        }
                    }
                    break;
                case TLzFieldType.lftSubquery:
                    TAlias alias1 = new TAlias();
                    alias1.column = field.AsText;
                    alias1.alias = field.AsText;
                    alias1.location = new Point(field.StartToken.XPosition, field.StartToken.YPosition);
                    alias1.tokenLine = field.StartToken.TokenLen;
                    if (field.aliasclause != null)
                    {
                        alias1.alias = field.FieldAlias;
                        alias1.column = field.FieldFullname;
                        alias1.location = new Point(field.aliasclause.AliasToken.XPosition, field.aliasclause.AliasToken.YPosition);
                        alias1.tokenLine = field.aliasclause.AliasToken.TokenLen;
                    }
                    if (level == 0)
                    {
                        aliases.Add(alias1);
                        if (!simply)
                        {
                            buffer.AppendLine("\r\nSearch " + alias1.alias + (level == 0 ? (" <<column_"
                                            + (++columnNumber) + ">>") : ""));
                        }
                    }
                    TSelectSqlStatement stmt = (TSelectSqlStatement)field.SubQuery;
                    List<TSelectSqlStatement> stmtList = new List<TSelectSqlStatement>();
                    getSelectSqlStatements(stmt, stmtList);
                    for (int i = 0; i < stmtList.Count; i++)
                    {
                        linkFieldToTables(alias1, stmtList[i].Fields[0], stmtList[i], level - 1);
                    }
                    break;
                default:
                    buffer.AppendLine("linkFieldToTables of type: " + field.FieldType + " not implemented yet");
                    break;
            }

            return ret;
        }

        private void getSelectSqlStatements(TSelectSqlStatement select,
        List<TSelectSqlStatement> stmtList)
        {
            if (select.SelectSetType != TSelectSetType.sltNone)
            {
                getSelectSqlStatements(select.LeftStmt, stmtList);
                getSelectSqlStatements(select.RightStmt, stmtList);
            }
            else
            {
                stmtList.Add(select);
            }
        }

        private void findColumnsFromGroupBy(TCustomSqlStatement select, int level)
        {
            if (select is TSelectSqlStatement && ((TSelectSqlStatement)select).GroupbyClause != null)
            {
                for (int j = 0; j < ((TSelectSqlStatement)select).GroupbyClause.GroupItems.Count(); j++)
                {
                    TLzGroupByItem i = (TLzGroupByItem)((TSelectSqlStatement)select).GroupbyClause.GroupItems[j];
                    List<TColumn> columns1 = exprToColumn((TLzCustomExpression)i._ndExpr, select, level);

                    TColumn column1 = columns1[0];
                    foreach (String tableName in column1.tableNames)
                    {
                        if (!simply)
                        {
                            buffer.AppendLine(new string(' ', level) + "--> " + column1.getFullName(tableName) + "(group by)");
                        }
                        findColumnInTables(column1, tableName, select, level + 1, ClauseType.groupby);
                    }

                }

            }
        }

        private bool isPseudocolumn(string column)
        {
            if (column == null)
                return false;
            if ("rownum".Equals(column.Trim().ToLower()))
                return true;
            else if ("rowid".Equals(column.Trim().ToLower()))
                return true;
            else if ("nextval".Equals(column.Trim().ToLower()))
                return true;
            else if ("sysdate".Equals(column.Trim().ToLower()))
                return true;
            return false;
        }

        private void findColumnsFromClauses(TCustomSqlStatement select, int level)
        {
            currentClauseMap.Add(select, ClauseType.undefine);
            Hashtable clauseTable = new Hashtable();
            List<TLzCustomExpression> keyList = new List<TLzCustomExpression>();
            if (select is TSelectSqlStatement)
            {

                TSelectSqlStatement statement = (TSelectSqlStatement)select;


                if (statement.SortClause != null)
                {
                    TLzOrderByList sortList = (TLzOrderByList)statement.SortClause;
                    for (int i = 0; i < sortList.Count(); i++)
                    {
                        TLzOrderBy orderBy = sortList[i];
                        TLzCustomExpression expr = orderBy.SortExpr;
                        clauseTable.Add(expr, ClauseType.orderby);
                        keyList.Add(expr);
                    }
                }


                if (statement.WhereClause != null)
                {
                    clauseTable.Add(statement.WhereClause, ClauseType.where);
                    keyList.Add(statement.WhereClause);
                }
                if (statement.ConnectByClause != null)
                {
                    clauseTable.Add(statement.ConnectByClause, ClauseType.connectby);
                    keyList.Add(statement.ConnectByClause);
                }
                if (statement.StartwithClause != null)
                {
                    clauseTable.Add(statement.StartwithClause, ClauseType.startwith);
                    keyList.Add(statement.StartwithClause);
                }
                if (statement.JoinTables != null)
                {
                    for (int i = 0; i < statement.JoinTables.Count(); i++)
                    {
                        TLzJoin join = statement.JoinTables[i];
                        if (join.JoinItems != null)
                        {
                            for (int j = 0; j < join.JoinItems.Count(); j++)
                            {
                                TLzJoinItem joinItem = join.JoinItems[j];
                                TLzCustomExpression expr = joinItem.JoinQual;
                                if (expr != null)
                                {
                                    clauseTable.Add(expr, ClauseType.join);
                                    keyList.Add(expr);
                                }
                            }
                        }
                    }
                }
            }
            else if (select is TUpdateSqlStatement)
            {
                TUpdateSqlStatement statement = (TUpdateSqlStatement)select;
                if (statement.SortClause != null)
                {
                    TLzOrderByList sortList = (TLzOrderByList)statement.SortClause;
                    for (int i = 0; i < sortList.Count(); i++)
                    {
                        TLzOrderBy orderBy = sortList[i];
                        TLzCustomExpression expr = orderBy.SortExpr;
                        clauseTable.Add(expr, ClauseType.orderby);
                        keyList.Add(expr);
                    }
                }
                if (statement.WhereClause != null)
                {
                    clauseTable.Add(statement.WhereClause, ClauseType.where);
                    keyList.Add(statement.WhereClause);
                }

                if (statement.JoinTables != null)
                {
                    for (int i = 0; i < statement.JoinTables.Count(); i++)
                    {
                        TLzJoin join = statement.JoinTables[i];
                        if (join.JoinItems != null)
                        {
                            for (int j = 0; j < join.JoinItems.Count(); j++)
                            {
                                TLzJoinItem joinItem = join.JoinItems[j];
                                TLzCustomExpression expr = joinItem.JoinQual;
                                if (expr != null)
                                {
                                    clauseTable.Add(expr, ClauseType.join);
                                    keyList.Add(expr);
                                }
                            }
                        }
                    }
                }
            }

            foreach (TLzCustomExpression expr in keyList)
            {
                currentClauseMap[select] = (ClauseType)clauseTable[expr];

                if (!simply)
                {
                    switch ((ClauseType)currentClauseMap[select])
                    {
                        case ClauseType.where:
                            buffer.AppendLine(new string(' ', level) + "--> Where Clause");
                            break;
                        case ClauseType.connectby:
                            buffer.AppendLine(new string(' ', level) + "--> Connect By Clause");
                            break;
                        case ClauseType.startwith:
                            buffer.AppendLine(new string(' ', level) + "--> Start With Clause");
                            break;
                        case ClauseType.orderby:
                            buffer.AppendLine(new string(' ', level) + "--> Order By Clause");
                            break;
                        case ClauseType.join:
                            buffer.AppendLine(new string(' ', level) + "--> Join");
                            break;
                    }

                }

                List<TColumn> columns = exprToColumn(expr, select, level);
                foreach (TColumn column1 in columns)
                {
                    foreach (String tableName in column1.tableNames)
                    {
                        if (!simply)
                        {

                            switch ((ClauseType)currentClauseMap[select])
                            {
                                case ClauseType.where:
                                    buffer.AppendLine(new string(' ', level + 1) + "--> " + column1.getFullName(tableName) + "(Where)");
                                    break;
                                case ClauseType.connectby:
                                    buffer.AppendLine(new string(' ', level + 1) + "--> " + column1.getFullName(tableName) + "(Connect By)");
                                    break;
                                case ClauseType.startwith:
                                    buffer.AppendLine(new string(' ', level + 1) + "--> " + column1.getFullName(tableName) + "(Start With)");
                                    break;
                                case ClauseType.orderby:
                                    buffer.AppendLine(new string(' ', level + 1) + "--> " + column1.getFullName(tableName) + "(Order By)");
                                    break;
                                case ClauseType.join:
                                    buffer.AppendLine(new string(' ', level + 1) + "--> " + column1.getFullName(tableName) + "(Join)");
                                    break;
                            }

                        }
                        findColumnInTables(column1, tableName, select, level + 2);
                    }

                }
            }
            currentClauseMap.Remove(select);

            // check order by clause
            findColumnsFromGroupBy(select, level);
        }

        private void findColumnsFromList(TCustomSqlStatement select, int level, TLz_List list)
        {
            if (list == null)
                return;

            for (int i = 0; i < list.Count(); i++)
            {
                TLzCustomExpression lcexpr = null;
                if (list[i] is TLz_SortGroupBy)
                {
                    lcexpr = (TLzCustomExpression)((TLz_SortGroupBy)list[i]).node;
                }
                else if (list[i] is TLzCustomExpression)
                {
                    lcexpr = (TLzCustomExpression)list[i];
                }

                if (lcexpr != null)
                {
                    List<TColumn> columns = exprToColumn(lcexpr, select, level);
                    foreach (TColumn column1 in columns)
                    {
                        findColumnInTables(column1, select, level + 1, null, null);
                        findColumnsFromClauses(select, level + 2);
                    }
                }
            }
        }

        internal TColumn attrToColumn(TLz_Attr lcexpr, TCustomSqlStatement stmt,
            TLzCustomExpression expr, bool collectExpr)
        {
            TColumn column = attrToColumn(lcexpr, stmt);
            if (collectExpr)
                column.expression = expr.AsText.Replace("\r\n", "\n")
                        .Replace("\n", " "); ;
            return column;
        }


        internal TColumn attrToColumn(TLz_Attr attr, TCustomSqlStatement stmt)
        {
            TColumn column = new TColumn();
            if (viewName != null)
                column.viewName = viewName;
            column.columnName = attr.LastToken.AsText;
            column.location = new Point((int)attr.LastToken.XPosition, (int)attr.LastToken.YPosition);
            column.tokenLine = attr.LastToken.TokenLen;
            if (attr.AsText.IndexOf(".") > 0)
            {
                column.columnPrex = attr.AsText.Substring(0, attr.AsText.Length - column.columnName.Length - ".".Length);
                String tableName = column.columnPrex;
                if (tableName.IndexOf(".") > 0)
                {
                    tableName = tableName.Substring(tableName.LastIndexOf(".") + 1);
                }
                if (!column.tableNames.Contains(tableName))
                {
                    column.tableNames.Add(tableName);
                }
            }
            else
            {
                TLzTableList tables = stmt.Tables;
                foreach (TLzTable lztable in tables)
                {
                    TTable table = TLzTaleToTable(lztable);
                    if (!column.tableNames.Contains(table.tableName))
                    {
                        column.tableNames.Add(table.tableName);
                    }
                }
            }

            column.orignColumn = column.columnName;

            return column;
        }

        TTable TLzTaleToTable(TLzTable lztable)
        {
            TTable table = new TTable();
            if (lztable.TableAttr != null && lztable.TableAttr.LastToken != null)
            {
                table.tableName = lztable.TableAttr.LastToken.AsText;
                if (lztable.TableAttr.AsText.IndexOf(".") > 0)
                {
                    table.prefixName = lztable.TableAttr.AsText.Substring(0, lztable.TableAttr.AsText.Length - table.tableName.Length - ".".Length);
                }
            }

            if (lztable.AliasClause != null)
            {
                table.tableAlias = lztable.AliasClause.AsText;
            }

            return table;

        }


        List<TColumn> exprToColumn(TLzCustomExpression expr, TCustomSqlStatement stmt, int level)
        {
            List<TColumn> columns = new List<TColumn>();

            columnsInExpr c = new columnsInExpr(this, expr, columns, stmt, level,
                false);
            c.searchColumn();

            return columns;
        }

        List<TColumn> exprToColumn(TLzCustomExpression expr, TCustomSqlStatement stmt, int level, bool collectExpr)
        {
            List<TColumn> columns = new List<TColumn>();

            columnsInExpr c = new columnsInExpr(this, expr, columns, stmt, level,
                collectExpr);
            c.searchColumn();

            return columns;
        }

        private bool isTopLevel(TLz_Node stmt)
	    {
		    TLz_Node parent = stmt.parentNode;
            if (parent is TSelectSqlStatement)
            {
                TSelectSqlStatement select = (TSelectSqlStatement)parent;
                if (select.SelectSetType != TSelectSetType.sltNone)
                {
                    return isTopLevel(select);
                }
                else
                    return false;
            }
            else if (parent != null)
            {
                return isTopLevel(parent);
            }
            else return true;
			    
	}


    }//class columnImpact
}
